@@ -31,7 +31,8 @@ from partis.pyproj import (
   norm_dist_filename )
 
 from partis.utils import (
-  init_logging,
+  logging_parser_add, 
+  logging_parser_init,
   log_levels,
   getLogger,
   LogListHandler,
@@ -41,15 +42,8 @@ from partis.utils import (
 
 log = getLogger(__name__)
 
-from .tool import (
-  Tool )
-
-from .testing import (
-  Test,
-  TestSuite,
-  TestSuiteResults )
-
 from .utils import (
+  dump_file,
   detect_run_with_mpi,
   get_mpiexec,
   get_processes,
@@ -70,12 +64,11 @@ from partis.schema.serialize.yaml import (
   dump,
   dumps )
 
-from partis.nwl.tool_pkg.build import build as build_pkg
-
 from .results import (
   ToolResults )
 
 from .utils import (
+  get_dirs,
   get_mpiexec,
   get_processes,
   get_cpus_per_process,
@@ -87,6 +80,14 @@ from .utils import (
   get_jobid,
   get_jobname )
 
+from .load_tool import load_tool_wait
+
+from .tool import (
+  Tool )
+
+from .workflow import (
+  Workflow )
+
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 
 def argument_parser( ):
@@ -116,7 +117,6 @@ def argument_parser( ):
     # dest = 'tool_name',
     help = "Alias for tool_name" )
 
-
   parser.add_argument( "--inputs",
     type = str,
     default = None,
@@ -127,15 +127,20 @@ def argument_parser( ):
     action='store_true',
     help = "Generate template tool inputs file using default values instead of running tool." )
 
-  parser.add_argument( "-d", "--rundir",
+  parser.add_argument( "-s", "--startdir",
     type = str,
     default = None,
-    help = "Run directory, relative path for output files" )
+    help = "Starting directory, relative path for input files" )
 
   parser.add_argument( "-c", "--workdir",
     type = str,
     default = None,
-    help = "Working directory, relative path for input files" )
+    help = "Workflow directory, relative path for input files" )
+
+  parser.add_argument( "-d", "--rundir",
+    type = str,
+    default = None,
+    help = "Run directory, relative path for output files" )
 
   parser.add_argument( "-e", "--environ",
     type = str,
@@ -182,6 +187,7 @@ def argument_parser( ):
     action='store_true',
     help = "Print extra tool documentation" )
 
+
   # parser.add_argument( "--test",
   #   type = str,
   #   default = list(),
@@ -198,16 +204,6 @@ def argument_parser( ):
     action='store_true',
     help = "Clears run directory before starting job" )
 
-  parser.add_argument( "-l", "--log",
-    type = str,
-    default = None,
-    help = "Redirect output to the given log file" )
-
-  parser.add_argument( "-v", "--verbosity",
-    type = str,
-    default = 'info',
-    help = f"Log verbosity {log_levels}" )
-
   parser.add_argument( "--venv",
     type = str,
     default = None,
@@ -234,6 +230,14 @@ def argument_parser( ):
     help = """Search location to find installable tool packages. The current
       directory is automatically added.""" )
 
+  parser.add_argument( "--serial",
+    action = 'store_true',
+    help = """Preserves the initial order of workflow stages unless a dependency
+      between two stages (I/O query, or an explicit 'afterok') would be violated.
+      All stages are run sequentially (no parallel execution).""" )
+
+  logging_parser_add(parser)
+
   return parser
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@@ -252,163 +256,11 @@ def parse_key_values( raw_vars ):
 
   return vars
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def load_tool(
-  name,
-  logger,
-  venv,
-  find_links ):
-
-  tool_mod = None
-  deps_installed = False
-  tmpdir = None
-  _find_links = list()
-
-
-  if logger.isEnabledFor( logging.DEBUG ):
-    _verbosity = ['-v', '-v', '-v']
-  else:
-    _verbosity = list()
-
-  _pip_args = [
-    # the cache is not safe when changed from multiple processes
-    '--no-cache-dir',
-    # not going to upgrade pip here
-    '--disable-pip-version-check',
-    # no user interaction
-    '--no-input' ]
-
-
-  try:
-    if osp.isfile( name ):
-
-      tool_name = norm_dist_filename( osp.basename( name ).rsplit('.', 1)[0] )
-      _name = f"nwlpkg_{tool_name}.{tool_name}"
-
-      try:
-        with venv:
-          tool_mod = importlib.import_module(_name)
-
-        name = _name
-
-        tool = tool_mod.tool
-
-        results_schema = tool_mod.results.results
-
-      except ImportError:
-
-
-        logger.info(f'Building tool package: {name}')
-
-        tmpdir = tempfile.mkdtemp()
-
-        find_links.append( tmpdir )
-
-        import trio
-        from functools import partial
-
-        pkg_name = trio.run( partial( build_pkg,
-          tool_file = name,
-          outdir = tmpdir,
-          logger = logger,
-          build_docs = False ) )
-
-        name = f"{pkg_name}.{tool_name}"
-
-    logger.info(ModelHint(
-      f'PIP Links',
-      hints = find_links ) )
-
-    for f in ['.'] + find_links:
-      logger.info(f"  '{f}'")
-
-      _find_links.extend([
-        '--find-links',
-        f ])
-
-    try:
-      with venv:
-        tool_mod = importlib.import_module(name)
-
-      tool = tool_mod.tool
-
-      results_schema = tool_mod.results.results
-
-    except ImportError:
-      pass
-
-    if not tool_mod:
-
-      # search current directory for a distribution containing the tool
-      pkg, _, mod = name.partition('.')
-
-      logger.info(f"Installing tool package: {pkg}")
-
-      try:
-
-        venv.install([
-            *_verbosity,
-            *_find_links,
-            *_pip_args,
-            f'{pkg}[run]' ],
-          env = {
-            **os.environ,
-            'PIP_NO_BUILD_ISOLATION' : 'True' })
-
-      except subprocess.CalledProcessError:
-        logger.error(
-          f"Failed to install tool package: {pkg}")
-
-        return None, None
-
-      deps_installed = True
-
-      with venv:
-        tool_mod = importlib.import_module(name)
-
-      tool = tool_mod.tool
-
-      results_schema = tool_mod.results.results
-
-    tool_dir = osp.dirname( osp.abspath(tool_mod.__file__) )
-    data_dir = osp.join( tool_dir, os.pardir, 'data' )
-
-    # update tool resource data paths
-    for name, data in tool.resources.static.items():
-      data_path = data.path
-
-      if not osp.isabs( data_path ):
-        tool.resources.static[name] = osp.join( data_dir, data_path )
-
-    if not deps_installed and len(tool.resources.python.dependencies) > 0:
-      # ensure all tool dependencies are installed
-      try:
-        venv.install([
-            '--upgrade',
-            *_verbosity,
-            *_find_links,
-            *_pip_args,
-            *[ str(dep) for dep in tool.resources.python.dependencies ] ],
-          env = {
-            **os.environ,
-            'PIP_NO_BUILD_ISOLATION' : 'True' })
-
-      except subprocess.CalledProcessError:
-        logger.error(
-          f"Failed to install tool dependencies: {pkg}")
-
-        return None, None
-
-  finally:
-    if tmpdir:
-      shutil.rmtree( tmpdir )
-
-  return tool, results_schema
-
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#
 def dump_error_results(
   log,
   log_handler,
+  startdir,
   workdir,
   rundir,
   aux ):
@@ -417,7 +269,9 @@ def dump_error_results(
   if not osp.exists( rundir ):
     os.makedirs(rundir)
 
-  results_file = osp.join( rundir, "nwl.tool.results.yml" )
+  results_file = osp.join( rundir, "nwl.results.yml" )
+  log_file = osp.join( rundir, "nwl.log.yml" )
+  env_file = osp.join( rundir, "nwl.env.yml")
 
   results = ToolResults()
 
@@ -429,6 +283,7 @@ def dump_error_results(
   results.job.host = get_jobhost()
   results.job.user = get_jobuser()
 
+  results.runtime.startdir = startdir
   results.runtime.workdir = workdir
   results.runtime.rundir = rundir
   results.runtime.aux = aux
@@ -439,18 +294,28 @@ def dump_error_results(
     re.sub( r'[^A-Za-z0-9\_]+', "_", k ) : v
     for k,v in env.items() }
 
-  results.runtime.env = env
-
-  results.runtime.logs = log_handler.logs
-
   try:
 
     try:
 
-      dump(
-        results_file,
-        results,
-        add_hash = True )
+      dump_file(
+        obj = env,
+        fname = env_file,
+        num = 0,
+        log = log.error )
+
+      dump_file(
+        obj = log_handler.logs,
+        fname = log_file,
+        num = 0,
+        log = log.error )
+
+      dump_file(
+        obj = results,
+        fname = results_file,
+        num = 0,
+        log = log.error )
+
 
     except Exception as e:
       log.error( ModelHint(
@@ -476,15 +341,22 @@ def positional_defaults(*args, **kwargs):
   return kwargs
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def run_tool( log, tool_log, args ):
+def run_tool( log, nwl_log, args ):
 
   if os.getcwd() not in sys.path:
     sys.path.append(os.getcwd())
 
+  pwd = osp.realpath( os.getcwd() )
+
   # keep complete record of runner log in case tool fails
   log_handler = LogListHandler()
   log.addHandler( log_handler )
 
+  venv_log = nwl_log.getChild("venv")
+  venv_log.propagate = False
+  venv_log_handler = LogListHandler()
+  venv_log.addHandler( venv_log_handler )
+
   # extract from either positional or named arguments
   tool_name, inputs_file = positional_defaults(
     args.tool_name,
@@ -492,24 +364,83 @@ def run_tool( log, tool_log, args ):
     tool_name = args.tool,
     inputs_file = args.inputs ).values()
 
-  pwd = osp.abspath( os.getcwd() )
-  workdir = args.workdir
-  rundir = args.rundir
+  #.............................................................................
+  # If given a previous result file, instead of tool name, use the result file
+  # to fill in all arguments as a re-run
+  if osp.basename(tool_name).startswith("nwl.results.yml"):
+
+    if not osp.exists(tool_name):
+      log.error(
+        f"File not found: {tool_name}")
+
+      return 1
+
+    tool_name = osp.realpath(tool_name)
+
+    if not ( args.truncate or (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) ):
+      log.error(
+        f"Set ``--truncate`` to over-write previous result: {tool_name}")
+
+      return 1
+
+    elif not args.truncate:
+      args.truncate = input(f"Overwrite (--truncate) previous results [yes/no]? ").lower().startswith('y')
+
+    if not args.truncate:
+      log.error(
+        f"Run cancelled by user.")
+
+      return 1
+
+    dir = args.rundir or osp.dirname(tool_name)
+
+    results_file = osp.join(dir, "nwl.results.yml")
+
+    if osp.exists(results_file):
+
+      results_file_bak = osp.join(
+        dir,
+        f"nwl.results.yml.bak_{int(time.time())}" )
+
+      log.info(f"Saving backup: {results_file_bak}")
+      shutil.copy2(results_file, results_file_bak)
+
+    log.info(f'Loading from previous run result: {tool_name}')
+
+    prev_results = load(
+      tool_name,
+      schema = ToolResults )
+
+    tool_name = prev_results.job.tool_qualname
+
+    if inputs_file is None:
+      inputs_file = prev_results.job.inputs_file
+
+      if not osp.abspath(inputs_file) and pwd != prev_results.job.curdir:
+        inputs_file = osp.join( prev_results.job.curdir, inputs_file )
+
+    args.startdir = args.startdir or prev_results.runtime.startdir
+    args.workdir = args.workdir or prev_results.runtime.workdir
+    args.rundir = args.rundir or prev_results.runtime.rundir
+    args.mpiexec = args.mpiexec or prev_results.runtime.mpiexec
+    args.np = args.np or prev_results.runtime.processes
+    args.ncp = args.ncp or prev_results.runtime.cpus_per_process
+    args.ntc = args.ntc or prev_results.runtime.threads_per_cpu
+    args.ngp = args.ngp or prev_results.runtime.gpus_per_process
+
+  pwd, startdir, workdir, rundir = get_dirs(
+    tool_name = tool_name,
+    startdir = args.startdir,
+    workdir = args.workdir,
+    rundir = args.rundir )
+
+
+  #.............................................................................
   venv_dir = args.venv
 
-  if not workdir:
-    workdir = pwd
-
-  if not rundir:
-    rundir = workdir
-
   if not venv_dir:
     venv_dir = osp.join( rundir, 'venv_nwlrun' )
 
-  workdir = osp.abspath( workdir )
-
-  rundir = osp.abspath( rundir )
-
   venv_dir = osp.abspath( venv_dir )
 
   find_links = [
@@ -529,16 +460,14 @@ def run_tool( log, tool_log, args ):
     dump_error_results(
       log = log,
       log_handler = log_handler,
+      startdir = startdir,
       workdir = workdir,
       rundir = rundir,
       aux = aux )
 
     return 1
 
-  log.info(f'Tool name: {tool_name}')
-  log.info(f'Inputs file: {inputs_file}')
-  log.info(f'Working dir: {workdir}')
-  log.info(f'Run dir: {rundir}')
+  log.info(f'Run: {tool_name}')
   log.info(f'Venv dir: {venv_dir}')
 
   try:
@@ -575,32 +504,36 @@ def run_tool( log, tool_log, args ):
     venv_in = args.venv_in
 
     venv_mutex = MutexFile(
-        prefix = osp.basename(venv_dir),
-        dir = osp.dirname(venv_dir),
-        # 10 min
-        timeout = 600.0 )
+      prefix = osp.basename(venv_dir),
+      dir = osp.dirname(venv_dir),
+      # 10 min
+      timeout = 600.0 )
 
     venv = VirtualEnv(
       path = venv_dir,
       inherit_site_packages = venv_in,
       reuse_existing = not args.venv_force,
       args = ['--without-pip'],
-      logger = log,
+      logger = venv_log,
       mutex = venv_mutex )
 
-    tool, results_schema = load_tool(
+    tool, results_schema = load_tool_wait(
       name = tool_name,
-      logger = log,
       venv = venv,
       find_links = find_links )
 
-    if not tool and ( args.venv_in or not args.venv_force ):
+    if tool is None and ( args.venv_in or not args.venv_force ):
       # install failed. Attempt again, but with no re-use or inheritance
       # from other environments
       # NOTE: non-install errors will have raised an exception instead of
       # returning None
       # NOTE: existing install(s) may lead to incompatible dependencies
-      log.warning(f"Re-initializing virtual environment without inhertied 'site-packages'")
+      log.warning(ModelHint(
+        f"Failed to load tool, re-attempting without inherited site-packages",
+        hints = venv_log_handler.hints ))
+
+      venv_log_handler.clear()
+
       venv_in = list()
 
       venv = VirtualEnv(
@@ -608,19 +541,25 @@ def run_tool( log, tool_log, args ):
         inherit_site_packages = venv_in,
         reuse_existing = False,
         args = ['--without-pip'],
-        logger = log,
+        logger = venv_log,
         mutex = venv_mutex )
 
-      tool, results_schema = load_tool(
+      tool, results_schema = load_tool_wait(
         name = tool_name,
-        logger = log,
         venv = venv,
         find_links = find_links )
 
-    if not tool:
+    if tool is None:
+      # still could not load tool
       returncode = 1
 
-    if returncode == 0:
+      log.error(ModelHint(
+        f"Failed to load tool",
+        hints = [ModelHint.from_dict(kw) for kw in venv_log_handler.logs] ))
+
+      venv_log_handler.clear()
+
+    if returncode == 0 and tool.type == 'workflow':
       with venv:
         # NOTE: using venv as context will update 'os.environ['PATH'] and sys.path
         # to approximate the environment seen by an interpreter run in the venv
@@ -631,7 +570,38 @@ def run_tool( log, tool_log, args ):
 
         schema_plugins.load_plugins()
 
-        log.info(f'Loaded tool: {tool_name}')
+        log.success(f'Loaded workflow: {tool_name}')
+
+        result = tool.run_wait(
+          startdir = startdir,
+          workdir = workdir,
+          rundir = rundir,
+          workflow_file = tool_name,
+          venv = venv,
+          log = nwl_log.getChild("workflow"),
+          initlogs = log_handler.logs,
+          find_links = find_links,
+          processes = args.np,
+          cpus_per_process = args.ncp,
+          threads_per_cpu = args.ntc,
+          gpus_per_process = args.ngp,
+          run_serial = args.serial )
+
+        tool_closed = True
+        returncode = 1 if tool._failed else 0
+
+    if returncode == 0 and tool.type == 'tool':
+      with venv:
+        # NOTE: using venv as context will update 'os.environ['PATH'] and sys.path
+        # to approximate the environment seen by an interpreter run in the venv
+
+        # ensure plugs reloaded with access to additional search paths
+        from partis.schema.plugin import (
+          schema_plugins )
+
+        schema_plugins.load_plugins()
+
+        log.success(f'Loaded tool: {tool_name}')
 
         if args.doc:
           log.info( "Input Schema" )
@@ -644,16 +614,20 @@ def run_tool( log, tool_log, args ):
         if args.template:
           init_val = results_schema.data.inputs.schema.init_val
 
-          log.info( dumps(init_val) )
-
           if inputs_file:
             dump( inputs_file, init_val )
 
             log.info(f"Tool inputs template generated: {inputs_file}")
 
+          else:
+            log.info( ModelHint(
+              "Tool inputs template",
+              data = dumps(init_val) ))
+
         elif inputs_file is not None:
           # run tool
           result = tool.run_wait(
+            startdir = startdir,
             workdir = workdir,
             rundir = rundir,
             env = env,
@@ -667,9 +641,9 @@ def run_tool( log, tool_log, args ):
             threads_per_cpu = args.ntc,
             gpus_per_process = args.ngp,
             inputs_file = inputs_file,
-            jobargs = sys.argv,
+            jobargs = list(sys.argv),
             results_schema = results_schema,
-            log = tool_log,
+            log = nwl_log.getChild("tool"),
             initlogs = log_handler.logs )
 
           tool_closed = True
@@ -680,7 +654,7 @@ def run_tool( log, tool_log, args ):
 
     returncode = 1
     log.error( ModelHint(
-      "Tool runtime failure",
+      "Runtime failure",
       hints = e ) )
 
   if inputs_file is not None and not ( args.template or tool_closed ):
@@ -688,6 +662,7 @@ def run_tool( log, tool_log, args ):
     dump_error_results(
       log = log,
       log_handler = log_handler,
+      startdir = startdir,
       workdir = workdir,
       rundir = rundir,
       aux = aux )
@@ -695,60 +670,14 @@ def run_tool( log, tool_log, args ):
   return returncode
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-# def run_tests():
-#
-#   if returncode == 0 and len(args.test) > 0:
-#
-#     try:
-#       test_suite = TestSuite()
-#
-#       for i, ftest in enumerate( args.test ):
-#         with open( ftest, 'r' ) as fp:
-#           doc = fp.read()
-#
-#         test_suite.tests[ str(i) ] = loads(
-#           src = doc,
-#           schema = TestSuite,
-#           loc = Loc(
-#             filename = ftest ) )
-#
-#       if len( test_suite.tests ) > 0:
-#         test_results = test_suite._run_wait(
-#           logger = log,
-#           workdir = args.workdir,
-#           rundir = args.rundir )
-#
-#         log.info(f"Tests: {test_results.npass} passed, {test_results.nfail} failed")
-#
-#         if not test_results.success:
-#           returncode = 1
-#
-#           # TODO: give the option to save results somewhere
-#           failed = list()
-#           stack = list( test_results.tests.items() )
-#
-#           while len(stack) > 0:
-#             path, test = stack.pop(0)
-#
-#             if not test.success:
-#               if isinstance( test, TestSuiteResults ):
-#                 for k, _test in test.tests.items():
-#                   stack.append( (f"{path}.{k}", _test) )
-#
-#               else:
-#                 failed.append( ( path, test ) )
-#
-#           for path, test in failed:
-#             log.info(f"{path}\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++")
-#             for l in test.logs:
-#               log.log( logging.getLevelName(l.level), l.msg )
-#             log.info(f"{path}\n------------------------------------------------------------------\n")
-#
-#     except Exception as e:
-#       returncode = 1
-#       log.exception( "Tool test failure" )
-#
-#   return returncode
+def feat_enabled(enabled, disabled):
+  if not ( enabled or disabled ):
+    return None
+
+  if enabled:
+    return True
+
+  return False
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 def main():
@@ -756,12 +685,16 @@ def main():
   parser = argument_parser( )
   args = parser.parse_args( )
 
-  init_logging(
-    level = args.verbosity,
-    filename = args.log )
+  logging_parser_init(args)
 
-  tool_log = getLogger(f"nwl")
-  log = tool_log.getChild("run")
+  # TODO: find better way to disable logging from other packages.
+  # Maybe instead setLevel to individual packages/loggers in init_logging
+  # instead of setting it on the 'root' logger for all packages
+  for l in ['matplotlib', 'numba', 'trio', 'trimesh']:
+    getLogger(l).setLevel(logging.ERROR)
+
+  nwl_log = getLogger(f"nwl")
+  log = nwl_log.getChild("run")
 
   if detect_run_with_mpi():
     log.error(
@@ -771,7 +704,7 @@ def main():
 
   time_start = timer()
 
-  returncode = run_tool( log, tool_log, args )
+  returncode = run_tool( log, nwl_log, args )
 
   time_end = timer()
   walltime = time_end - time_start
@@ -779,7 +712,7 @@ def main():
   if returncode != 0:
     log.error(f"Job exited with errors: wall-time {datetime.timedelta(seconds = walltime)} (H:M:S)")
   else:
-    log.info(f"Job completed successfully: wall-time {datetime.timedelta(seconds = walltime)} (H:M:S)")
+    log.success(f"Job completed successfully: wall-time {datetime.timedelta(seconds = walltime)} (H:M:S)")
 
   return returncode
 
