@@ -1,6 +1,7 @@
 # -*- coding: UTF-8 -*-
 
 import os
+from functools import wraps
 
 import logging
 log = logging.getLogger(__name__)
@@ -18,6 +19,7 @@ from partis.schema import (
   derived,
   is_optional,
   PyEvaluated,
+  PyEvaluatedRestricted,
   PassPrim,
   BoolPrim,
   IntPrim,
@@ -48,23 +50,26 @@ from .query import (
   NWLQueryEvaluated )
 
 from .context import (
-  InputsContext )
+  QueryContext,
+  EnabledInputContext )
 
 from .content_type import (
   ContentTypeList )
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-EvaluatedInputs = PyEvaluated.subclass(
-  context = InputsContext )
+# evaluates input queries to get values from other tools/files
+EvaluatedInputQuery = NWLQueryEvaluated.subclass(
+  context = QueryContext )
 
-EvaluatedInputValues = NWLQueryEvaluated.subclass(
-  context = InputsContext )
+# evaluates the 'enabled' expressions on inputs
+EvaluatedInputEnabled = PyEvaluatedRestricted.subclass(
+  context = EnabledInputContext )
 
 WorkAbsolutePathEvaluatedInput = WorkAbsolutePathEvaluated.subclass(
-  context = InputsContext )
+  context = QueryContext )
 
 RunAbsolutePathEvaluatedInput = RunAbsolutePathEvaluated.subclass(
-  context = InputsContext )
+  context = QueryContext )
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 # create declareds for all the 'input' schemas
@@ -94,8 +99,8 @@ AnyNonUnionInput = UnionPrim(
     dict_declared,
     struct_declared,
     file_declared,
-    wfile_declared,
     dir_declared,
+    wfile_declared,
     wdir_declared ],
   default_case = 0 )
 
@@ -112,11 +117,25 @@ AnyInput = UnionPrim(
     struct_declared,
     union_declared,
     file_declared,
-    wfile_declared,
     dir_declared,
+    wfile_declared,
     wdir_declared ],
   default_case = 0 )
 
+#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+def set_schema_origin(f):
+
+  @wraps(f)
+  def _f(self, *args, **kwargs):
+    val = f(self, *args, **kwargs)
+
+    if val is not None:
+      val.schema_origin = self
+
+    return val
+
+  return _f
+
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 def process_base( schema ):
 
@@ -126,7 +145,11 @@ def process_base( schema ):
   if schema.doc is not None:
     doc.append(schema.doc)
 
-  if hasattr(schema, 'selection') and schema.selection is not None and len(schema.selection) > 0:
+  if hasattr(schema, 'selection') and schema.selection is not None:
+    if len(schema.selection) == 0:
+      raise SchemaError(
+        f"`selection` must be non-empty list")
+
     restricted = list()
 
     for s in schema.selection:
@@ -177,9 +200,10 @@ class BaseInput( StructValued ):
   enabled = BoolPrim(
     doc = "Marks the input as enabled if True, disabled if False",
     default_val = True,
-    evaluated = EvaluatedInputs )
+    evaluated = EvaluatedInputEnabled )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     """Returns the schema for the value defined by this input
 
@@ -229,6 +253,7 @@ class BoolInput( BaseInput ):
     default_val = False )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -236,7 +261,7 @@ class BoolInput( BaseInput ):
       default_val = self.default_val,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -280,6 +305,7 @@ class IntInput( BaseInput ):
     default_val = optional )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -290,7 +316,7 @@ class IntInput( BaseInput ):
       max = self.max,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -334,6 +360,7 @@ class FloatInput( BaseInput ):
     default_val = optional )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -344,7 +371,7 @@ class FloatInput( BaseInput ):
       max = self.max,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -426,6 +453,7 @@ class StrInput( BaseInput ):
     default_val = optional )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -440,7 +468,7 @@ class StrInput( BaseInput ):
       strip = self.strip,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -466,6 +494,7 @@ class ListInput( BaseInput ):
     default_val = list() )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -477,7 +506,7 @@ class ListInput( BaseInput ):
       min_len = self.min_len,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -502,6 +531,7 @@ class DictInput( BaseInput ):
     default_val = dict() )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -513,7 +543,7 @@ class DictInput( BaseInput ):
       min_len = self.min_len,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -541,6 +571,7 @@ class StructInput( BaseInput ):
     default_val = optional )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -566,7 +597,7 @@ class StructInput( BaseInput ):
       default_val = default_val,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -609,6 +640,7 @@ class UnionInput( BaseInput ):
     max_cols = 80 )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -637,7 +669,7 @@ class UnionInput( BaseInput ):
       default_case = default_case,
       doc = doc,
       module = module,
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc )
 
@@ -661,6 +693,10 @@ class PathValue( StructValued ):
 
   path = StrPrim( )
 
+  #-----------------------------------------------------------------------------
+  def __fspath__(self):
+    return self.path
+
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class PathInputValue( PathValue ):
   pass
@@ -680,9 +716,8 @@ class PathInput( BaseInput ):
     doc = "If true, requires the path to be non-empty",
     default_val = False )
 
-  content_types = ContentTypeList
-
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     raise NotImplementedError("PathInputValue sublcass must implement")
 
@@ -693,7 +728,10 @@ class WorkFileInput( PathInput ):
   schema = dict(
     declared = file_declared )
 
+  content_types = ContentTypeList
+
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -702,7 +740,7 @@ class WorkFileInput( PathInput ):
       doc = doc,
       module = module,
       # NOTE: this is for evaluating a query
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc,
       default_val = derived,
@@ -723,7 +761,10 @@ class RunFileInput( PathInput ):
   schema = dict(
     declared = wfile_declared )
 
+  content_types = ContentTypeList
+
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -732,7 +773,7 @@ class RunFileInput( PathInput ):
       doc = doc,
       module = module,
       # NOTE: this is for evaluating a query
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc,
       default_val = derived,
@@ -753,6 +794,7 @@ class WorkDirInput( PathInput ):
     declared = dir_declared )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -761,7 +803,7 @@ class WorkDirInput( PathInput ):
       doc = doc,
       module = module,
       # NOTE: this is for evaluating a query
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc,
       default_val = derived,
@@ -782,6 +824,7 @@ class RunDirInput( PathInput ):
     declared = wdir_declared )
 
   #-----------------------------------------------------------------------------
+  @set_schema_origin
   def value_schema( self, name, tag = None, module = None ):
     doc, restricted = process_base( self )
 
@@ -790,7 +833,7 @@ class RunDirInput( PathInput ):
       doc = doc,
       module = module,
       # NOTE: this is for evaluating a query
-      evaluated = EvaluatedInputValues,
+      evaluated = EvaluatedInputQuery,
       name = name,
       loc = self._loc,
       default_val = derived,
