@@ -4,6 +4,7 @@ import logging
 log = logging.getLogger(__name__)
 
 from partis.schema import (
+  SchemaError,
   required,
   optional,
   derived,
@@ -114,7 +115,7 @@ class BaseOutput( StructValued ):
 
   #-----------------------------------------------------------------------------
   def value_schema( self, name, tag = None, module = None ):
-    raise SchemaError(f"Value schema not implemented for this output")
+    raise NotImplementedError(f"Value schema not implemented for this output")
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class BoolOutput( BaseOutput ):
@@ -555,6 +556,10 @@ class PathOutputValue( StructValued ):
     # evaluated = EvaluatedOutputs | RunRelativePathEvaluated,
     max_lines = 1 )
 
+  #-----------------------------------------------------------------------------
+  def __fspath__(self):
+    return self.path
+
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class PathOutput( BaseOutput ):
   schema = dict(
