@@ -23,6 +23,9 @@ from partis.utils import (
   indent_lines,
   LogListHandler )
 
+from partis.utils.inspect import (
+  filter_traceback )
+
 from partis.schema import (
   required,
   optional,
@@ -46,7 +49,9 @@ from partis.schema import (
   schema_declared,
   SchemaHint,
   SchemaError,
-  Loc,
+  Loc )
+
+from partis.schema.hint import (
   Hint,
   HintList )
 
@@ -55,8 +60,15 @@ from partis.schema.prim.any_prim import (
   AnyPrim )
 
 from .log import (
+  LogContext,
   LogEvent )
 
+from .base import (
+  test_declared,
+  test_result_declared,
+  test_suite_declared,
+  test_suite_result_declared )
+
 from .tool import (
   Tool )
 
@@ -77,46 +89,22 @@ class NWLTestHint( ModelHint ):
   """
   pass
 
-
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-test_declared = schema_declared( tag = 'test' )
-test_result_declared = schema_declared( tag = 'test' )
-
-test_suite_declared = schema_declared( tag = 'test_suite' )
-test_suite_result_declared = schema_declared( tag = 'test_suite' )
-
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-async def subprocess_log_cmd( args, cwd, logger ):
-
-  run_ret = await trio.run_process(
-    command = args,
-    capture_stdout = True,
-    capture_stderr = True,
-    check = False )
-
-  logger.info( "run stdout\n" + indent_lines(2, run_ret.stdout.decode("ascii", errors = "replace") ) )
-  logger.info( "run stderr\n" + indent_lines(2, run_ret.stderr.decode("ascii", errors = "replace") ) )
-
-  if run_ret.returncode != 0:
-
-    raise subprocess.CalledProcessError(
-      run_ret.returncode,
-      run_ret.args,
-      output = run_ret.stdout,
-      stderr = run_ret.stderr )
-
-  return run_ret
-
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 def diff_hint( path, ref, val ):
   _path = ".".join(path)
 
   if is_sequence(ref):
     if not is_sequence( val ):
-      return SchemaHint(f"expected sequence at `{_path}`: {fmt_base_or_type(val)}")
+      return SchemaHint(
+        f"Expected a sequence",
+        data = val,
+        loc = Loc(path = path) )
 
     if len( val ) != len( ref ):
-      return SchemaHint(f"sequence at `{_path}` length {len(ref)}: {len(val)}")
+      return SchemaHint(
+        f"Expected a sequence of length {len(ref)}",
+        data = len(val),
+        loc = Loc(path = path) )
 
     hints = list()
 
@@ -135,7 +123,10 @@ def diff_hint( path, ref, val ):
 
   elif is_mapping( ref ):
     if not is_mapping( val ):
-      return SchemaHint(f"expected mapping at `{_path}`: {fmt_base_or_type(val)}")
+      return SchemaHint(
+        f"Expected mapping",
+        data = val,
+        loc = Loc(path = path) )
 
     # if len( val ) != len( ref ):
     #   return SchemaHint(f"mapping at `{_path}` length {len(ref)} : {len(val)}")
@@ -148,7 +139,7 @@ def diff_hint( path, ref, val ):
       _hint = None
 
       if k not in val:
-        _hint = SchemaHint(f"missing key: `{k}`")
+        _hint = SchemaHint(f"Expected key: `{k}`")
 
       else:
 
@@ -411,7 +402,11 @@ class Test( StructValued ):
     default_val = list() )
 
   #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-  async def _run( self, logger, workdir, rundir ):
+  async def _run( self,
+    log,
+    venv,
+    workdir,
+    rundir ):
 
     if workdir is None:
       workdir = os.getcwd()
@@ -423,7 +418,7 @@ class Test( StructValued ):
     rundir = os.path.normpath( rundir )
 
     log_handler = LogListHandler()
-    logger.addHandler( log_handler )
+    log.addHandler( log_handler )
 
     if os.path.exists( rundir ):
       shutil.rmtree( rundir )
@@ -458,7 +453,7 @@ class Test( StructValued ):
         loc = Loc(
           filename = tool_file ) )
 
-      tool_results_file = os.path.join( rundir, "nwl.tool.results.yml" )
+      tool_results_file = os.path.join( rundir, "nwl.results.yml" )
 
       in_file = os.path.join( rundir, "tool_inputs.yml" )
 
@@ -475,26 +470,73 @@ class Test( StructValued ):
       _mpiexec = " ".join(self.mpiexec)
 
       # run the tool with given inputs
-      await subprocess_log_cmd( [
-          "partis-nwl",
-          "--tool",
-          tool_file,
-          "--inputs",
-          in_file,
-          "--workdir",
-          workdir,
-          "--rundir",
-          rundir,
-          "--np",
-          str(self.processes),
-          "--ncp",
-          str(self.cpus_per_process),
-          "--ntc",
-          str(self.threads_per_cpu),
-          "--mpiexec",
-          f"\"{_mpiexec}\"" ],
-        cwd = workdir,
-        logger = logger )
+      args = [
+        "partis-nwl",
+        "--tool",
+        tool_file,
+        "--inputs",
+        in_file,
+        "--workdir",
+        workdir,
+        "--rundir",
+        rundir,
+        "--np",
+        str(self.processes),
+        "--ncp",
+        str(self.cpus_per_process),
+        "--ntc",
+        str(self.threads_per_cpu),
+        "--mpiexec",
+        f"\"{_mpiexec}\"" ]
+
+
+      log.info(ModelHint(
+        f'Starting test',
+        hints = [
+          ModelHint(
+            'Config',
+            level = 'info',
+            hints = {
+              'tool' : tool_file,
+              'rundir': rundir,
+              'inputs_file': in_file } ),
+          ModelHint(
+            'Run Arguments',
+            level = 'debug',
+            hints = args ) ] ))
+
+      res = await venv.trio_run(
+        args,
+        capture_stdout = True,
+        stderr = subprocess.STDOUT,
+        check = False )
+
+      out = res.stdout.decode('utf-8', errors = 'replace')
+
+      if res.returncode == 0:
+        log.success(f'Finished: {rundir}')
+
+      else:
+        lines = out.splitlines()[-50:]
+        err_hint = None
+
+        if len( lines ) > 0:
+          txt = "\n".join(lines)
+
+          err_hint = ModelHint(
+            f"last {len( lines )} lines of output",
+            level = 'warning',
+            data = txt,
+            format = 'block' )
+
+        log.error(ModelHint(
+          f'Tool error return code: {res.returncode}',
+          hints = err_hint ))
+
+        raise subprocess.CalledProcessError(
+          res.returncode,
+          res.args,
+          output = res.stdout )
 
       with open( tool_results_file, "r" ) as fp:
         doc = fp.read()
@@ -508,7 +550,7 @@ class Test( StructValued ):
 
 
       if not tool_results.runtime.success:
-        logger.error( f"Tool runtime failure" )
+        log.error( f"Tool runtime failure" )
 
       else:
 
@@ -520,7 +562,7 @@ class Test( StructValued ):
             val = tool_results.data.commands )
 
           if hint is not None:
-            logger.error( hint )
+            log.error( hint )
 
         # check that outputs have correct values
         if self.outputs is not None:
@@ -530,7 +572,7 @@ class Test( StructValued ):
             val = tool_results.data.outputs )
 
           if hint is not None:
-            logger.error( hint )
+            log.error( hint )
 
 
         # check that output files have correct values
@@ -539,31 +581,37 @@ class Test( StructValued ):
             file._rfile( dir = rundir )
 
           except Exception as e:
-            logger.error( NWLTestHint(
+            log.error( NWLTestHint(
               f"Output file error",
               level = 'error',
               hints = e ) )
 
         # check test logs
-        logs = self.logs._eval(
-          locals = tool_results,
-          locals_key = '_',
-          context = "log" )
+        with filter_traceback(
+          suppress = True,
+          log = log.error,
+          msg = f"Test log evaluation failed" ):
 
-        # filter for only enabled log events
-        for l in logs:
-          if l.enabled:
-            logger.log( logging.getLevelName(l.level), l.msg )
+          logs = self.logs._eval(
+            context = LogContext(
+              results = tool_results,
+              static = None ),
+            logger = log )
+
+          # filter for only enabled log events
+          for l in logs:
+            if l.enabled:
+              log.log( logging.getLevelName(l.level), l.msg )
 
     except Exception as e:
-      logger.error( NWLTestHint(
+      log.error( NWLTestHint(
         f"Test failed due to exception",
         level = 'error',
         hints = e ) )
 
 
     results.logs = log_handler.logs
-    logger.removeHandler( log_handler )
+    log.removeHandler( log_handler )
 
     results.success = not any(
       l['level'] in [ 'ERROR', 'CRITICAL' ]
@@ -572,8 +620,13 @@ class Test( StructValued ):
     return results
 
   #-----------------------------------------------------------------------------
-  def _run_wait( self, logger, workdir, rundir ):
-    results = trio.run( self._run, logger, workdir, rundir )
+  def _run_wait( self,
+    log,
+    venv,
+    workdir,
+    rundir ):
+
+    results = trio.run( self._run, log, venv, workdir, rundir )
     return results
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@@ -632,7 +685,7 @@ class TestSuite( StructValued ):
     default_val = dict() )
 
   #-----------------------------------------------------------------------------
-  async def _run( self, logger, workdir, rundir, propagate = False ):
+  async def _run( self, log, venv, workdir, rundir, propagate = False ):
 
     if workdir is None:
       workdir = os.getcwd()
@@ -653,11 +706,12 @@ class TestSuite( StructValued ):
       name = f"test_{k}"
 
       _rundir = os.path.join( rundir, name )
-      _logger = logger.getChild( name )
-      _logger.propagate = propagate
+      _log = log.getChild( name )
+      _log.propagate = propagate
 
       runs[k] = test._run(
-        logger = _logger,
+        log = _log,
+        venv = venv,
         workdir = workdir,
         rundir = _rundir )
 
@@ -685,6 +739,6 @@ class TestSuite( StructValued ):
     return results
 
   #-----------------------------------------------------------------------------
-  def _run_wait( self, logger, workdir, rundir ):
-    results = trio.run( self._run, logger, workdir, rundir )
+  def _run_wait( self, log, venv, workdir, rundir ):
+    results = trio.run( self._run, log, venv, workdir, rundir )
     return results
