@@ -8,11 +8,11 @@ import shlex
 from timeit import default_timer as timer
 import socket
 from string import Formatter
-
+import signal
 import logging
-log = logging.getLogger(__name__)
 
 from partis.utils import (
+  getLogger,
   odict,
   adict,
   ModelHint,
@@ -20,6 +20,14 @@ from partis.utils import (
   LogListHandler,
   ProcessEnv )
 
+from partis.utils.inspect import (
+  filter_traceback )
+
+from partis.utils.sig import (
+  listenable,
+  add_signal_listener,
+  remove_signal_listener )
+
 from partis.schema import (
   required,
   optional,
@@ -46,16 +54,23 @@ from partis.schema import (
   schema_declared,
   SchemaModule )
 
+from partis.schema.hint import (
+  Hint,
+  HintList )
+
 from .base import (
+  tool_declared,
   ToolError )
 
 from .inputs import (
   AnyInput,
-  InputsContext )
+  QueryContext,
+  EnabledInputContext )
 
 from .commands import (
   AnyCommand,
-  CommandsContext )
+  CommandsContext,
+  ProcessCommand )
 
 from .outputs import (
   AnyOutput,
@@ -70,11 +85,10 @@ from partis.schema.serialize.yaml import (
   load,
   dump )
 
-from .utils import (
-  check_inout_files )
-
 from .info import ToolInfo
 
+from .runtime import Env, EnabledInput
+
 from .resources import ToolResources
 
 from .results import (
@@ -82,6 +96,8 @@ from .results import (
   ToolResults )
 
 from .utils import (
+  dump_file,
+  get_dirs,
   get_mpiexec,
   get_processes,
   get_nodelist,
@@ -92,13 +108,17 @@ from .utils import (
   get_jobhost,
   get_jobuser,
   get_jobid,
-  get_jobname )
+  get_jobname,
+  get_input_query_deps,
+  get_inputs_enabled,
+  filter_inputs_enabled,
+  checked_inout_files )
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class Tool( StructValued ):
   schema = dict(
-    tag = 'tool',
     doc = "Tool definition",
+    declared = tool_declared,
     default_val = derived )
 
   nwl = IntPrim(
@@ -163,6 +183,7 @@ class Tool( StructValued ):
     self._opened = False
     self._closed = False
     self._failed = False
+    self._save_num = 0
 
   #-----------------------------------------------------------------------------
   def tool_doc( self ):
@@ -289,8 +310,62 @@ class Tool( StructValued ):
 
     return results_schema
 
+
+  #-----------------------------------------------------------------------------
+  def _update_result_logs( self ):
+    if self._log_handler is None:
+      return
+
+    self._results_log.extend(self._log_handler.logs)
+    self._log_handler.clear()
+
+    # command fails if any prolog event of 'ERROR' or 'CRITICAL'
+    self._failed = self._failed or any(
+      l['level'] in [ 'ERROR', 'CRITICAL' ]
+      for l in self._results_log )
+
+  #-----------------------------------------------------------------------------
+  def _save_results( self ):
+    if not self._opened:
+      raise ValueError("Tool never opened")
+
+    dump_file(
+      obj = self._results,
+      fname = self._results_file,
+      num = self._save_num,
+      log = self._runtime_log.error,
+      # don't remove defaults for results file so that it can be read without
+      # needing to know the schema
+      no_defaults = False )
+
+    dump_file(
+      obj = self._results_log,
+      fname = self._log_file,
+      num = self._save_num,
+      log = self._runtime_log.error )
+
+    self._save_num += 1
+
+  #-----------------------------------------------------------------------------
+  def _signal_listener( self, sig ):
+    name = sig.name
+    signals = self._results.runtime.signals
+
+    if name not in signals:
+      signals[name] = 0
+
+    signals[name] += 1
+
+    self._save_results()
+
+    if sig in [signal.SIGTERM, signal.SIGINT]:
+      if isinstance(self._cur_cmd, ProcessCommand):
+        if self._cur_cmd._process is not None:
+          self._cur_cmd._process.send_signal(sig)
+
   #-----------------------------------------------------------------------------
   def _config( self, *,
+    startdir = None,
     workdir = None,
     rundir = None,
     env = None,
@@ -382,6 +457,7 @@ class Tool( StructValued ):
       Pre-generated results schema to use
     initlogs : list
       Logs to add to results that occurred before running tool
+
     Returns
     -------
     None
@@ -393,9 +469,9 @@ class Tool( StructValued ):
     self._configured = True
 
     if log is not None:
-      self._log = log.getChild(f"tool")
+      self._log = log
     else:
-      self._log = logging.getLogger(f"tool")
+      self._log = getLogger(f"nwl.tool")
 
     self._log_handler = LogListHandler()
     self._log.addHandler( self._log_handler )
@@ -408,20 +484,16 @@ class Tool( StructValued ):
     # issue non-command logs to child logger to prvent mixing
     self._runtime_log = self._log.getChild("runtime")
 
-    self._pwd = osp.realpath( os.getcwd() )
+    self._pwd, self._startdir, self._workdir, self._rundir = get_dirs(
+      tool_name = self.qualname,
+      startdir = startdir,
+      workdir = workdir,
+      rundir = rundir )
 
-    if workdir is None:
-      workdir = self._pwd
-
-    if rundir is None:
-      rundir = workdir
-
-    workdir = osp.realpath( workdir )
-
-    rundir = osp.realpath( rundir )
-
-    self._workdir = workdir
-    self._rundir = rundir
+    self._log.info(f'Starting dir: {self._startdir}')
+    self._log.info(f'Working dir: {self._workdir}')
+    self._log.info(f'Run dir: {self._rundir}')
+    self._log.info(f'Inputs: {inputs_file}')
 
     if truncate is None:
       truncate = False
@@ -431,9 +503,6 @@ class Tool( StructValued ):
     self._timeout = timeout
 
 
-    self._results_file = osp.join( self._rundir, "nwl.tool.results.yml")
-    self._results_file_tmp = self._results_file + ".tmp"
-
     self._nodefile = osp.join( self._rundir, "nwl.nodes")
 
     if results_schema is None:
@@ -441,9 +510,55 @@ class Tool( StructValued ):
 
     # create a default results object
     self._results = results_schema()
+    self._results_file = osp.join( self._rundir, "nwl.results.yml")
 
+    self._results_log = HintList()
+    self._log_file = osp.join( self._rundir, "nwl.log.yml")
+
+    self._env = Env()
+    self._env_file = osp.join( self._rundir, "nwl.env.yml")
+
+    #.............................................................................
+    with filter_traceback(
+      suppress = True,
+      filter = Exception,
+      log = self._runtime_log.error,
+      msg = f"Tool inputs loading failed" ):
+
+      inputs_schema = self._results.data.inputs._schema
+
+      if not ( inputs or inputs_file ):
+        # default inputs
+        inputs = inputs_schema.decode( inputs_schema.init_val )
+
+      elif inputs_file:
+        # load from file
+        inputs = load(
+          file = inputs_file,
+          schema = inputs_schema,
+          loc = inputs_file )
+
+      else:
+        # validate
+        inputs = inputs_schema.decode(
+          inputs,
+          inputs_file )
+
+      self._results.job.inputs_file = inputs_file
+      self._results.data.inputs = inputs
+
+      self._results.runtime.query_deps = get_input_query_deps(
+        startdir = self._startdir,
+        workdir = self._workdir,
+        rundir = self._rundir,
+        name_path = ['data', 'inputs'],
+        val = inputs )
+
+      self._runtime_log.success(f"Inputs validated.")
+
+    #.............................................................................
     if initlogs:
-      self._results.runtime.logs += initlogs
+      self._results_log.extend( initlogs )
 
     self._results.job.tool_qualname = self.qualname
 
@@ -457,6 +572,7 @@ class Tool( StructValued ):
     self._results.job.curdir = self._pwd
 
     # set initial runtime information
+    self._results.runtime.startdir = self._startdir
     self._results.runtime.workdir = self._workdir
     self._results.runtime.rundir = self._rundir
 
@@ -469,7 +585,7 @@ class Tool( StructValued ):
         re.sub( r'[^A-Za-z0-9\_]+', "_", k ) : v
         for k,v in env.items() }
 
-      self._results.runtime.env = env
+      self._env.update(env)
 
     if aux is not None:
       self._results.runtime.aux = aux
@@ -547,8 +663,8 @@ class Tool( StructValued ):
         self._results.runtime.mpiexec = _mpiexec
 
     if self._results.runtime.threads_per_process > 1:
-      if 'OMP_NUM_THREADS' not in self._results.runtime.env:
-        self._results.runtime.env['OMP_NUM_THREADS'] = str(self._results.runtime.threads_per_process)
+      if 'OMP_NUM_THREADS' not in self._env:
+        self._env['OMP_NUM_THREADS'] = str(self._results.runtime.threads_per_process)
 
     if (
       not self.resources.multi_thread
@@ -597,114 +713,8 @@ class Tool( StructValued ):
 
     self._venv = venv
 
-    try:
-      inputs_schema = self._results.data.inputs._schema
-
-      if not ( inputs or inputs_file ):
-        # default inputs
-        inputs = inputs_schema.decode( inputs_schema.init_val )
-
-      elif inputs_file:
-        # load from file
-        inputs = load(
-          file = inputs_file,
-          schema = inputs_schema,
-          loc = inputs_file )
-
-      else:
-        # validate
-        inputs = inputs_schema.decode(
-          inputs,
-          inputs_file )
-
-      self._runtime_log.info(f"Inputs validated.")
-
-      try:
-        self._results.data.inputs = inputs._eval(
-          context = InputsContext(
-            results = self._results,
-            static = self.resources.static ),
-          logger = self._runtime_log.getChild("inputs") )
-
-        try:
-          logs = self.prolog._eval(
-            context = LogContext(
-              results = self._results,
-              static = self.resources.static ),
-            logger = self._runtime_log.getChild("prolog") )
-
-          # filter for only enabled log events
-          for l in logs:
-            if l.enabled:
-              self._runtime_log.log( logging.getLevelName(l.level), l.msg )
-
-          missing_files = check_inout_files(
-            dir = self._workdir,
-            path = [ 'data', 'inputs', ],
-            val = self._results.data.inputs )
-
-          if len( missing_files ) > 0:
-            self._runtime_log.warning( ModelHint(
-              msg = f"Input files not found: {len(missing_files)}",
-              hints = missing_files ) )
-
-        except Exception as e:
-          self._runtime_log.error( ModelHint(
-            msg = f"Tool prolog evaluation failed",
-            hints = e ) )
-
-      except Exception as e:
-        self._runtime_log.error( ModelHint(
-          msg = f"Tool inputs evaluation failed",
-          hints = e ) )
-
-    except Exception as e:
-      self._runtime_log.error( ModelHint(
-        msg = f"Tool inputs loading failed",
-        hints = e ) )
-
     self._update_result_logs()
 
-  #-----------------------------------------------------------------------------
-  def _update_result_logs( self ):
-    if self._log_handler is None:
-      return
-
-    self._results.runtime.logs += self._log_handler.logs
-    self._log_handler.clear()
-
-    # command fails if any prolog event of 'ERROR' or 'CRITICAL'
-    self._failed = self._failed or any(
-      l['level'] in [ 'ERROR', 'CRITICAL' ]
-      for l in self._results.runtime.logs )
-
-  #-----------------------------------------------------------------------------
-  def _save_results( self ):
-    if not self._opened:
-      raise ValueError("Tool never opened")
-
-    try:
-
-      try:
-        dump(
-          self._results_file_tmp,
-          self._results,
-          add_hash = True )
-
-        os.replace(
-          self._results_file_tmp,
-          self._results_file )
-
-      except Exception as e:
-        self._runtime_log.error( ModelHint(
-          msg = f"Failed to write tool result file",
-          hints = e ) )
-
-    except Exception as e:
-      self._runtime_log.error( ModelHint(
-        msg = f"Failed to encode tool result document",
-        hints = e ) )
-
   #-----------------------------------------------------------------------------
   def _open( self ):
     """Begin tool execution
@@ -722,27 +732,54 @@ class Tool( StructValued ):
 
     self._opened = True
 
-    try:
+    if self._failed:
+      return
+
+    with filter_traceback(
+      suppress = True,
+      filter = Exception,
+      log = self._runtime_log.error,
+      msg = f"Failed to prepare tool runtime directory: {self._rundir}" ):
+
       if not osp.exists( self._rundir ):
         os.makedirs( self._rundir )
 
       elif self._truncate:
-        shutil.rmtree( self._rundir )
-        os.mkdir( self._rundir )
+        # TODO: figure out what this should do. Simply deleting the rundir if it
+        # simply exists may cause problems, such as deleting the virtual environment
+        # that the tool was installed in.
+
+        # shutil.rmtree( self._rundir )
+        # os.mkdir( self._rundir )
+        pass
 
       if osp.exists( self._results_file ):
         os.remove( self._results_file )
 
-      os.chdir( self._rundir )
-
       if len(self._results.runtime.nodes) > 0:
         with open(self._nodefile, 'w') as fp:
           fp.write('\n'.join(self._results.runtime.nodes))
 
-    except Exception as e:
+      os.chdir( self._rundir )
+
+    dump_file(
+      obj = self._env,
+      fname = self._env_file,
+      num = 0,
+      log = self._runtime_log.error )
+
+    # check for file dependencies needed to evaluate input query expressions
+    missing_query_deps = [
+      f"{query_dep.name}: {shlex.quote(query_dep.path)}"
+      for query_dep in self._results.runtime.query_deps
+      # NOTE: 'tool' means 'this' results file, which doesn't exist yet and 
+      # is still in memory.
+      if query_dep.missing() and query_dep.base != 'tool' ]
+
+    if len(missing_query_deps) > 0:
       self._runtime_log.error( ModelHint(
-        msg = f"Failed to prepare tool runtime directory: {self._rundir}",
-        hints = e ) )
+        f"Missing input file dependencies: {len(missing_query_deps)}",
+        hints = missing_query_deps ))
 
     self._update_result_logs()
 
@@ -751,6 +788,73 @@ class Tool( StructValued ):
     if self._failed:
       return
 
+    with filter_traceback(
+      suppress = True,
+      log = self._runtime_log.error,
+      msg = f"Tool inputs evaluation failed" ):
+
+      # evaluate all queries in the inputs to get real values
+      self._results.data.inputs = self._results.data.inputs._eval(
+        context = QueryContext(
+          results = self._results,
+          static = self.resources.static ),
+        logger = self._runtime_log.getChild("inputs") )
+
+      # evaluate all 'enabled' expressions to determine which input values
+      # are enabled for given input combination
+      inputs_enabled = get_inputs_enabled(self._results.data.inputs)
+
+      inputs_enabled = inputs_enabled._eval(
+        context = EnabledInputContext(
+          results = self._results,
+          static = self.resources.static ),
+        logger = self._runtime_log.getChild("inputs") )
+
+      self._results.runtime.inputs_enabled = inputs_enabled
+
+      self._results.runtime.input_files = checked_inout_files(
+        dir = self._workdir,
+        name_path = [ 'data', 'inputs', ],
+        val = filter_inputs_enabled(
+          val = self._results.data.inputs,
+          inputs_enabled = inputs_enabled ) )
+
+      missing_files = [
+        f"{file_dep.name}: {shlex.quote(file_dep.path)}"
+        for file_dep in self._results.runtime.input_files
+        if file_dep.missing() ]
+
+      if len(missing_files) > 0:
+        self._runtime_log.error( ModelHint(
+          f"Missing input files: {len(missing_files)}",
+          hints = missing_files ))
+
+      with filter_traceback(
+        suppress = True,
+        log = self._runtime_log.error,
+        msg = f"Tool prolog evaluation failed" ):
+
+        logs = self.prolog._eval(
+          context = LogContext(
+            results = self._results,
+            static = self.resources.static ),
+          logger = self._runtime_log.getChild("prolog") )
+
+        # filter for only enabled log events
+        for l in logs:
+          if l.enabled:
+            self._runtime_log.log( logging.getLevelName(l.level), l.msg )
+
+    self._update_result_logs()
+
+    self._save_results()
+
+    if self._failed:
+      return
+
+    for sig in listenable:
+      add_signal_listener(sig, self._signal_listener)
+
     self._open_next_cmd()
 
   #-----------------------------------------------------------------------------
@@ -777,31 +881,33 @@ class Tool( StructValued ):
       self._cmd_handler.clear()
 
       try:
-        cmd = self.commands[ self._results.runtime.cmd_id ]
+        with filter_traceback(
+          suppress = False,
+          filter = SchemaError,
+          log = self._runtime_log.error,
+          msg = f"Command evaluation failed: `{self._results.runtime.cmd_id}`" ):
 
-        self._cur_cmd = cmd._eval(
-          context = CommandsContext(
-            results = self._results,
-            static = self.resources.static ),
-          logger = self._cmd_log.getChild( self._results.runtime.cmd_id ) )
+          cmd = self.commands[ self._results.runtime.cmd_id ]
 
-        self._results.data.commands[ self._results.runtime.cmd_id ].enabled = self._cur_cmd.enabled
+          self._cur_cmd = cmd._eval(
+            context = CommandsContext(
+              results = self._results,
+              static = self.resources.static ),
+            logger = self._cmd_log.getChild( self._results.runtime.cmd_id ) )
 
-        if not self._cur_cmd.enabled:
-          # non-enabled commands considered successful
-          self._results.data.commands[ self._results.runtime.cmd_id ].success = True
+          self._results.data.commands[ self._results.runtime.cmd_id ].enabled = self._cur_cmd.enabled
 
-          self._runtime_log.info(
-            f"Command skipped (not enabled): `{self._results.runtime.cmd_id}`")
+          if not self._cur_cmd.enabled:
+            # non-enabled commands considered successful
+            self._results.data.commands[ self._results.runtime.cmd_id ].success = True
 
-          # keep looping to find next enabled command
-          self._cur_cmd = None
+            self._runtime_log.info(
+              f"Command skipped (not enabled): `{self._results.runtime.cmd_id}`")
+
+            # keep looping to find next enabled command
+            self._cur_cmd = None
 
       except SchemaError as e:
-        self._runtime_log.error( ModelHint(
-          msg = f"Command evaluation failed: `{self._results.runtime.cmd_id}`",
-          hints = e ) )
-
         self._cur_cmd = None
         return
 
@@ -814,7 +920,7 @@ class Tool( StructValued ):
       tool = self,
       workdir = self._results.runtime.workdir,
       rundir = self._results.runtime.rundir,
-      env = self._results.runtime.env,
+      env = self._env,
       venv = self._venv,
       id = self._results.runtime.cmd_id,
       tool_results = self._results,
@@ -836,12 +942,17 @@ class Tool( StructValued ):
         self._cur_cmd._reset()
 
       self._cur_cmd = None
-      return
 
+
+    self._runtime_log.info(
+      f"Command starting: `{self._results.runtime.cmd_id}`")
+
+    self._update_result_logs()
     self._save_results()
 
-    self._cur_cmd._open(
-      tool_results = self._results )
+    if self._cur_cmd is not None:
+      self._cur_cmd._open(
+        tool_results = self._results )
 
   #-----------------------------------------------------------------------------
   def _poll( self ):
@@ -882,7 +993,7 @@ class Tool( StructValued ):
 
       return self._close()
 
-    self._runtime_log.info(
+    self._runtime_log.success(
       ModelHint(
         f"Command finished: `{self._results.runtime.cmd_id}`",
         hints = self._cmd_handler.hints ) )
@@ -923,17 +1034,18 @@ class Tool( StructValued ):
       eval_log.addHandler( eval_handler )
 
       try:
+        with filter_traceback():
 
-        self._results.data.outputs = self._results.data.outputs._eval(
-          context = OutputsContext(
-            results = self._results,
-            static = self.resources.static ),
-          logger = eval_log )
+          self._results.data.outputs = self._results.data.outputs._eval(
+            context = OutputsContext(
+              results = self._results,
+              static = self.resources.static ),
+            logger = eval_log )
 
-        self._runtime_log.info(
-          ModelHint(
-            f"Outputs evaluated",
-            hints = eval_handler.hints ) )
+          self._runtime_log.info(
+            ModelHint(
+              f"Outputs evaluated",
+              hints = eval_handler.hints ) )
 
       except Exception as e:
         self._runtime_log.error( ModelHint(
@@ -946,31 +1058,36 @@ class Tool( StructValued ):
         eval_log.removeHandler(eval_handler)
 
     if not self._failed:
-      missing_files = check_inout_files(
+      self._results.runtime.output_files = checked_inout_files(
         dir = self._rundir,
-        path = [ 'data', 'outputs', ],
+        name_path = [ 'data', 'outputs', ],
         val = self._results.data.outputs )
 
-      if len( missing_files ) > 0:
-        self._failed = True
+      missing_files = [
+        f"{file_dep.name}: {shlex.quote(file_dep.path)}"
+        for file_dep in self._results.runtime.output_files
+        if file_dep.missing() ]
 
+      if len(missing_files) > 0:
+        self._failed = True
         self._runtime_log.error( ModelHint(
-          msg = f"Output files not found: {len(missing_files)}",
-          hints = missing_files ) )
+          f"Missing output files: {len(missing_files)}",
+          hints = missing_files ))
 
     if not self._failed:
       # check epilogs
       try:
-        logs = self.epilog._eval(
-          context = LogContext(
-            results = self._results,
-            static = self.resources.static ),
-          logger = self._runtime_log.getChild("epilog") )
+        with filter_traceback():
+          logs = self.epilog._eval(
+            context = LogContext(
+              results = self._results,
+              static = self.resources.static ),
+            logger = self._runtime_log.getChild("epilog") )
 
-        # filter for only enabled log events
-        for l in logs:
-          if l.enabled:
-            self._runtime_log.log( logging.getLevelName(l.level), l.msg )
+          # filter for only enabled log events
+          for l in logs:
+            if l.enabled:
+              self._runtime_log.log( logging.getLevelName(l.level), l.msg )
 
       except Exception as e:
         self._runtime_log.error( ModelHint(
@@ -984,6 +1101,9 @@ class Tool( StructValued ):
 
     self._save_results()
 
+    for sig in listenable:
+      remove_signal_listener(sig, self._signal_listener)
+
     # return to initial working directory when finished
     os.chdir( self._pwd )
 
@@ -1024,8 +1144,10 @@ class Tool( StructValued ):
 
     hints = list()
 
-    hints.extend( results.data.inputs._lint(
-      context = InputsContext(
+    inputs_enabled = get_inputs_enabled(results.data.inputs)
+
+    hints.extend( inputs_enabled._lint(
+      context = EnabledInputContext(
         results = results,
         static = self.resources.static ) ) )
 
@@ -1039,7 +1161,7 @@ class Tool( StructValued ):
       hints.extend( cmd.lint( self, cmd_id, results ) )
 
 
-    hints.extend( results.data.outputs._lint(
+    hints.extend( self.outputs._lint(
       context = OutputsContext(
         results = results,
         static = self.resources.static ) ) )
