@@ -1,9 +1,7 @@
 
 from .tool_edit import ToolEditor
 from .tool_results_edit import ToolResultsEditor
-
-from .testing import TestSuiteWidget
-
+from .workflow_edit import WorkflowEditor
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 def get_view_editors():
@@ -19,7 +17,8 @@ def get_view_editors():
   plugins.append( EditorPluginGroup(
     editors = {
       "NWL Tool" : ToolEditor,
-      "NWL Tool Results" : _ToolResultsEditor } ) )
+      "NWL Tool Results" : _ToolResultsEditor,
+      "NWL Workflow" : WorkflowEditor } ) )
 
 
   return plugins
