@@ -27,7 +27,7 @@ from partis.schema import (
   is_sequence )
 
 from partis.nwl.context import (
-  InputsContext,
+  EnabledInputContext,
   OutputsContext,
   CommandsContext )
 
